"""
Generate a Python extension module with the constants defined in linux/input.h.
"""

import getopt
import os
import re
import sys

# -----------------------------------------------------------------------------
# The default header file locations to try.
headers = [
    "/usr/include/linux/input.h",
    "/usr/include/linux/input-event-codes.h",
    "/usr/include/linux/uinput.h",
]

opts, args = getopt.getopt(sys.argv[1:], "", ["ecodes", "stubs", "reproducible"])
if not opts:
    print("usage: genecodes.py [--ecodes|--stubs] [--reproducible] <headers>")
    exit(2)

if args:
    headers = args

reproducible = ("--reproducible", "") in opts


# -----------------------------------------------------------------------------
macro_regex = r"#define\s+((?:KEY|ABS|REL|SW|MSC|LED|BTN|REP|SND|ID|EV|BUS|SYN|FF|UI_FF|INPUT_PROP)_\w+)"
macro_regex = re.compile(macro_regex)

if reproducible:
    uname = "hidden for reproducibility"
else:
    # Uname without hostname.
    uname = list(os.uname())
    uname = " ".join((uname[0], *uname[2:]))


# -----------------------------------------------------------------------------
template_ecodes = r"""
#include <Python.h>
#ifdef __FreeBSD__
#include <dev/evdev/input.h>
#include <dev/evdev/uinput.h>
#else
#include <linux/input.h>
#include <linux/uinput.h>
#endif

/* Automatically generated by evdev.genecodes */
/* Generated on   %s */
/* Generated from %s */

#define MODULE_NAME "_ecodes"
#define MODULE_HELP "linux/input.h macros"

static PyMethodDef MethodTable[] = {
    { NULL, NULL, 0, NULL}
};

static struct PyModuleDef moduledef = {
    PyModuleDef_HEAD_INIT,
    MODULE_NAME,
    MODULE_HELP,
    -1,          /* m_size */
    MethodTable, /* m_methods */
    NULL,        /* m_reload */
    NULL,        /* m_traverse */
    NULL,        /* m_clear */
    NULL,        /* m_free */
};

PyMODINIT_FUNC
PyInit__ecodes(void)
{
    PyObject* m = PyModule_Create(&moduledef);
    if (m == NULL) return NULL;

%s

    return m;
}
"""


template_stubs = r"""
# Automatically generated by evdev.genecodes
# Generated on %s
# Generated from %s

# pylint: skip-file

ecodes: dict[str, int]
keys: dict[int, str|list[str]]
bytype: dict[int, dict[int, str|list[str]]]

KEY: dict[int, str|list[str]]
ABS: dict[int, str|list[str]]
REL: dict[int, str|list[str]]
SW:  dict[int, str|list[str]]
MSC: dict[int, str|list[str]]
LED: dict[int, str|list[str]]
BTN: dict[int, str|list[str]]
REP: dict[int, str|list[str]]
SND: dict[int, str|list[str]]
ID:  dict[int, str|list[str]]
EV:  dict[int, str|list[str]]
BUS: dict[int, str|list[str]]
SYN: dict[int, str|list[str]]
FF_STATUS:     dict[int, str|list[str]]
FF_INPUT_PROP: dict[int, str|list[str]]

%s
"""


def parse_headers(headers=headers):
    for header in headers:
        try:
            fh = open(header)
        except (IOError, OSError):
            continue

        for line in fh:
            macro = macro_regex.search(line)
            if macro:
                yield macro.group(1)


all_macros = list(parse_headers())
if not all_macros:
    print("no input macros found in: %s" % " ".join(headers), file=sys.stderr)
    sys.exit(1)

# pylint: disable=possibly-used-before-assignment, used-before-assignment
if ("--ecodes", "") in opts:
    body = ("    PyModule_AddIntMacro(m, %s);" % macro for macro in all_macros)
    template = template_ecodes
elif ("--stubs", "") in opts:
    body = ("%s: int" % macro for macro in all_macros)
    template = template_stubs

body = os.linesep.join(body)
text = template % (uname, headers if not reproducible else ["hidden for reproducibility"], body)
print(text.strip())



import sys
from unittest import mock
from pprint import PrettyPrinter

sys.modules["evdev.ecodes"] = mock.Mock()
from evdev import ecodes_runtime as ecodes

pprint = PrettyPrinter(indent=2, sort_dicts=True, width=120).pprint


print("# Automatically generated by evdev.genecodes_py")
print()
print('"""')
print(ecodes.__doc__.strip())
print('"""')

print()
print("from typing import Final, Dict, Tuple, Union")
print()

for name, value in ecodes.ecodes.items():
    print(f"{name}: Final[int] = {value}")
print()

entries = [
    ("ecodes", "Dict[str, int]", "#: Mapping of names to values."),
    ("bytype", "Dict[int, Dict[int, Union[str, Tuple[str]]]]", "#: Mapping of event types to other value/name mappings."),
    ("keys",   "Dict[int, Union[str, Tuple[str]]]", "#: Keys are a combination of all BTN and KEY codes."),
    ("KEY",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("ABS",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("REL",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("SW",     "Dict[int, Union[str, Tuple[str]]]", None),
    ("MSC",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("LED",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("BTN",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("REP",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("SND",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("ID",     "Dict[int, Union[str, Tuple[str]]]", None),
    ("EV",     "Dict[int, Union[str, Tuple[str]]]", None),
    ("BUS",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("SYN",    "Dict[int, Union[str, Tuple[str]]]", None),
    ("FF",     "Dict[int, Union[str, Tuple[str]]]", None),
    ("UI_FF",  "Dict[int, Union[str, Tuple[str]]]", None),
    ("FF_STATUS",  "Dict[int, Union[str, Tuple[str]]]", None),
    ("INPUT_PROP", "Dict[int, Union[str, Tuple[str]]]", None)
]

for key, annotation, doc in entries:
    if doc:
        print(doc)

    print(f"{key}: {annotation} = ", end="")
    pprint(getattr(ecodes, key))
    print()


import ctypes

from . import ecodes

_u8 = ctypes.c_uint8
_u16 = ctypes.c_uint16
_u32 = ctypes.c_uint32
_s16 = ctypes.c_int16
_s32 = ctypes.c_int32


class Replay(ctypes.Structure):
    """
    Defines scheduling of the force-feedback effect
    @length: duration of the effect
    @delay: delay before effect should start playing
    """

    _fields_ = [
        ("length", _u16),
        ("delay", _u16),
    ]


class Trigger(ctypes.Structure):
    """
    Defines what triggers the force-feedback effect
    @button: number of the button triggering the effect
    @interval: controls how soon the effect can be re-triggered
    """

    _fields_ = [
        ("button", _u16),
        ("interval", _u16),
    ]


class Envelope(ctypes.Structure):
    """
    Generic force-feedback effect envelope
    @attack_length: duration of the attack (ms)
    @attack_level: level at the beginning of the attack
    @fade_length: duration of fade (ms)
    @fade_level: level at the end of fade

    The @attack_level and @fade_level are absolute values; when applying
    envelope force-feedback core will convert to positive/negative
    value based on polarity of the default level of the effect.
    Valid range for the attack and fade levels is 0x0000 - 0x7fff
    """

    _fields_ = [
        ("attack_length", _u16),
        ("attack_level", _u16),
        ("fade_length", _u16),
        ("fade_level", _u16),
    ]


class Constant(ctypes.Structure):
    """
    Defines parameters of a constant force-feedback effect
    @level: strength of the effect; may be negative
    @envelope: envelope data
    """

    _fields_ = [
        ("level", _s16),
        ("ff_envelope", Envelope),
    ]


class Ramp(ctypes.Structure):
    """
    Defines parameters of a ramp force-feedback effect
    @start_level: beginning strength of the effect; may be negative
    @end_level: final strength of the effect; may be negative
    @envelope: envelope data
    """

    _fields_ = [
        ("start_level", _s16),
        ("end_level", _s16),
        ("ff_envelope", Envelope),
    ]


class Condition(ctypes.Structure):
    """
    Defines a spring or friction force-feedback effect
    @right_saturation: maximum level when joystick moved all way to the right
    @left_saturation: same for the left side
    @right_coeff: controls how fast the force grows when the joystick moves to the right
    @left_coeff: same for the left side
    @deadband: size of the dead zone, where no force is produced
    @center: position of the dead zone
    """

    _fields_ = [
        ("right_saturation", _u16),
        ("left_saturation", _u16),
        ("right_coeff", _s16),
        ("left_coeff", _s16),
        ("deadband", _u16),
        ("center", _s16),
    ]


class Periodic(ctypes.Structure):
    """
    Defines parameters of a periodic force-feedback effect
    @waveform: kind of the effect (wave)
    @period: period of the wave (ms)
    @magnitude: peak value
    @offset: mean value of the wave (roughly)
    @phase: 'horizontal' shift
    @envelope: envelope data
    @custom_len: number of samples (FF_CUSTOM only)
    @custom_data: buffer of samples (FF_CUSTOM only)
    """

    _fields_ = [
        ("waveform", _u16),
        ("period", _u16),
        ("magnitude", _s16),
        ("offset", _s16),
        ("phase", _u16),
        ("envelope", Envelope),
        ("custom_len", _u32),
        ("custom_data", ctypes.POINTER(_s16)),
    ]


class Rumble(ctypes.Structure):
    """
    Defines parameters of a periodic force-feedback effect
    @strong_magnitude: magnitude of the heavy motor
    @weak_magnitude: magnitude of the light one

    Some rumble pads have two motors of different weight. Strong_magnitude
    represents the magnitude of the vibration generated by the heavy one.
    """

    _fields_ = [
        ("strong_magnitude", _u16),
        ("weak_magnitude", _u16),
    ]


class EffectType(ctypes.Union):
    _fields_ = [
        ("ff_constant_effect", Constant),
        ("ff_ramp_effect", Ramp),
        ("ff_periodic_effect", Periodic),
        ("ff_condition_effect", Condition * 2),  # one for each axis
        ("ff_rumble_effect", Rumble),
    ]


class Effect(ctypes.Structure):
    _fields_ = [
        ("type", _u16),
        ("id", _s16),
        ("direction", _u16),
        ("ff_trigger", Trigger),
        ("ff_replay", Replay),
        ("u", EffectType),
    ]


class UInputUpload(ctypes.Structure):
    _fields_ = [
        ("request_id", _u32),
        ("retval", _s32),
        ("effect", Effect),
        ("old", Effect),
    ]


class UInputErase(ctypes.Structure):
    _fields_ = [
        ("request_id", _u32),
        ("retval", _s32),
        ("effect_id", _u32),
    ]


# ff_types = {
#     ecodes.FF_CONSTANT,
#     ecodes.FF_PERIODIC,
#     ecodes.FF_RAMP,
#     ecodes.FF_SPRING,
#     ecodes.FF_FRICTION,
#     ecodes.FF_DAMPER,
#     ecodes.FF_RUMBLE,
#     ecodes.FF_INERTIA,
#     ecodes.FF_CUSTOM,
# }



import collections
import glob
import os
import re
import stat
from typing import Union, List

from . import ecodes
from .events import InputEvent, event_factory, KeyEvent, RelEvent, AbsEvent, SynEvent


def list_devices(input_device_dir: Union[str, bytes, os.PathLike] = "/dev/input") -> List[str]:
    """List readable character devices in ``input_device_dir``."""

    fns = glob.glob("{}/event*".format(input_device_dir))
    return list(filter(is_device, fns))


def is_device(fn: Union[str, bytes, os.PathLike]) -> bool:
    """Check if ``fn`` is a readable and writable character device."""

    if not os.path.exists(fn):
        return False

    m = os.stat(fn)[stat.ST_MODE]
    if not stat.S_ISCHR(m):
        return False

    if not os.access(fn, os.R_OK | os.W_OK):
        return False

    return True


def categorize(event: InputEvent) -> Union[InputEvent, KeyEvent, RelEvent, AbsEvent, SynEvent]:
    """
    Categorize an event according to its type.

    The :data:`event_factory <evdev.events.event_factory>` dictionary
    maps event types to sub-classes of :class:`InputEvent
    <evdev.events.InputEvent>`. If the event cannot be categorized, it
    is returned unmodified."""

    if event.type in event_factory:
        return event_factory[event.type](event)
    else:
        return event


def resolve_ecodes_dict(typecodemap, unknown="?"):
    """
    Resolve event codes and types to their verbose names.

    :param typecodemap: mapping of event types to lists of event codes.
    :param unknown: symbol to which unknown types or codes will be resolved.

    Example
    -------
    >>> resolve_ecodes_dict({ 1: [272, 273, 274] })
    { ('EV_KEY', 1): [('BTN_MOUSE',  272),
                      ('BTN_RIGHT',  273),
                      ('BTN_MIDDLE', 274)] }

    If ``typecodemap`` contains absolute axis info (instances of
    :class:`AbsInfo <evdev.device.AbsInfo>` ) the result would look
    like:

    >>> resolve_ecodes_dict({ 3: [(0, AbsInfo(...))] })
    { ('EV_ABS', 3L): [(('ABS_X', 0L), AbsInfo(...))] }
    """

    for etype, codes in typecodemap.items():
        type_name = ecodes.EV[etype]

        # ecodes.keys are a combination of KEY_ and BTN_ codes
        if etype == ecodes.EV_KEY:
            ecode_dict = ecodes.keys
        else:
            ecode_dict = getattr(ecodes, type_name.split("_")[-1])

        resolved = resolve_ecodes(ecode_dict, codes, unknown)
        yield (type_name, etype), resolved


def resolve_ecodes(ecode_dict, ecode_list, unknown="?"):
    """
    Resolve event codes and types to their verbose names.

    Example
    -------
    >>> resolve_ecodes(ecodes.BTN, [272, 273, 274])
    [(['BTN_LEFT', 'BTN_MOUSE'], 272), ('BTN_RIGHT', 273), ('BTN_MIDDLE', 274)]
    """
    res = []
    for ecode in ecode_list:
        # elements with AbsInfo(), eg { 3 : [(0, AbsInfo(...)), (1, AbsInfo(...))] }
        if isinstance(ecode, tuple):
            if ecode[0] in ecode_dict:
                l = ((ecode_dict[ecode[0]], ecode[0]), ecode[1])
            else:
                l = ((unknown, ecode[0]), ecode[1])

        # just ecodes, e.g: { 0 : [0, 1, 3], 1 : [30, 48] }
        else:
            if ecode in ecode_dict:
                l = (ecode_dict[ecode], ecode)
            else:
                l = (unknown, ecode)
        res.append(l)

    return res


def find_ecodes_by_regex(regex):
    """
    Find ecodes matching a regex and return a mapping of event type to event codes.

    regex can be a pattern string or a compiled regular expression object.

    Example
    -------
    >>> find_ecodes_by_regex(r'(ABS|KEY)_BR(AKE|EAK)')
    {1: [411], 3: [10]}
    >>> res = find_ecodes_by_regex(r'(ABS|KEY)_BR(AKE|EAK)')
    >>> resolve_ecodes_dict(res)
    {
        ('EV_KEY', 1): [('KEY_BREAK', 411)],
        ('EV_ABS', 3): [('ABS_BRAKE', 10)]
    }
    """

    regex = re.compile(regex)  # re.compile is idempotent
    result = collections.defaultdict(list)

    for type_code, codes in ecodes.bytype.items():
        for code, names in codes.items():
            names = (names,) if isinstance(names, str) else names
            for name in names:
                if regex.match(name):
                    result[type_code].append(code)
                    break

    return dict(result)


__all__ = ("list_devices", "is_device", "categorize", "resolve_ecodes", "resolve_ecodes_dict", "find_ecodes_by_regex")



import contextlib
import os
from typing import Dict, Generic, Iterator, List, Literal, NamedTuple, Tuple, TypeVar, Union, overload

from . import _input, ecodes, util

try:
    from .eventio_async import EvdevError, EventIO
except ImportError:
    from .eventio import EvdevError, EventIO

_AnyStr = TypeVar("_AnyStr", str, bytes)


class AbsInfo(NamedTuple):
    """Absolute axis information.

    A ``namedtuple`` with absolute axis information -
    corresponds to the ``input_absinfo`` struct:

    Attributes
    ---------
    value
      Latest reported value for the axis.

    min
      Specifies minimum value for the axis.

    max
      Specifies maximum value for the axis.

    fuzz
      Specifies fuzz value that is used to filter noise from the
      event stream.

    flat
      Values that are within this value will be discarded by joydev
      interface and reported as 0 instead.

    resolution
      Specifies resolution for the values reported for the axis.
      Resolution for main axes (``ABS_X, ABS_Y, ABS_Z``) is reported
      in units per millimeter (units/mm), resolution for rotational
      axes (``ABS_RX, ABS_RY, ABS_RZ``) is reported in units per
      radian.

    Note
    ----
    The input core does not clamp reported values to the ``[minimum,
    maximum]`` limits, such task is left to userspace.

    """

    value: int
    min: int
    max: int
    fuzz: int
    flat: int
    resolution: int

    def __str__(self):
        return "value {}, min {}, max {}, fuzz {}, flat {}, res {}".format(*self)  # pylint: disable=not-an-iterable


class KbdInfo(NamedTuple):
    """Keyboard repeat rate.

    Attributes
    ----------
    delay
      Amount of time that a key must be depressed before it will start
      to repeat (in milliseconds).

    repeat
      Keyboard repeat rate in characters per second.
    """

    delay: int
    repeat: int

    def __str__(self):
        return "delay {}, repeat {}".format(self.delay, self.repeat)


class DeviceInfo(NamedTuple):
    """
    Attributes
    ----------
    bustype
    vendor
    product
    version
    """

    bustype: int
    vendor: int
    product: int
    version: int

    def __str__(self) -> str:
        msg = "bus: {:04x}, vendor {:04x}, product {:04x}, version {:04x}"
        return msg.format(*self)  # pylint: disable=not-an-iterable


class InputDevice(EventIO, Generic[_AnyStr]):
    """
    A linux input device from which input events can be read.
    """

    __slots__ = ("path", "fd", "info", "name", "phys", "uniq", "_rawcapabilities", "version", "ff_effects_count")

    def __init__(self, dev: Union[_AnyStr, "os.PathLike[_AnyStr]"]):
        """
        Arguments
        ---------
        dev : str|bytes|PathLike
          Path to input device
        """

        #: Path to input device.
        self.path: _AnyStr = dev if not hasattr(dev, "__fspath__") else dev.__fspath__()

        # Certain operations are possible only when the device is opened in read-write mode.
        try:
            fd = os.open(dev, os.O_RDWR | os.O_NONBLOCK)
        except OSError:
            fd = os.open(dev, os.O_RDONLY | os.O_NONBLOCK)

        #: A non-blocking file descriptor to the device file.
        self.fd: int = fd

        # Returns (bustype, vendor, product, version, name, phys, capabilities).
        info_res = _input.ioctl_devinfo(self.fd)

        #: A :class:`DeviceInfo <evdev.device.DeviceInfo>` instance.
        self.info = DeviceInfo(*info_res[:4])

        #: The name of the event device.
        self.name: str = info_res[4]

        #: The physical topology of the device.
        self.phys: str = info_res[5]

        #: The unique identifier of the device.
        self.uniq: str = info_res[6]

        #: The evdev protocol version.
        self.version: int = _input.ioctl_EVIOCGVERSION(self.fd)

        #: The raw dictionary of device capabilities - see `:func:capabilities()`.
        self._rawcapabilities = _input.ioctl_capabilities(self.fd)

        #: The number of force feedback effects the device can keep in its memory.
        self.ff_effects_count = _input.ioctl_EVIOCGEFFECTS(self.fd)

    def __del__(self) -> None:
        if hasattr(self, "fd") and self.fd is not None:
            try:
                self.close()
            except (OSError, ImportError, AttributeError):
                pass

    def _capabilities(self, absinfo: bool = True):
        res = {}

        for etype, _ecodes in self._rawcapabilities.items():
            for code in _ecodes:
                l = res.setdefault(etype, [])
                if isinstance(code, tuple):
                    if absinfo:
                        a = code[1]  # (0, 0, 0, 255, 0, 0)
                        i = AbsInfo(*a)
                        l.append((code[0], i))
                    else:
                        l.append(code[0])
                else:
                    l.append(code)

        return res

    @overload
    def capabilities(self, verbose: Literal[False] = ..., absinfo: bool = ...) -> Dict[int, List[int]]:
        ...
    @overload
    def capabilities(self, verbose: Literal[True], absinfo: bool = ...) -> Dict[Tuple[str, int], List[Tuple[str, int]]]:
        ...
    def capabilities(self, verbose: bool = False, absinfo: bool = True) -> Union[Dict[int, List[int]], Dict[Tuple[str, int], List[Tuple[str, int]]]]:
        """
        Return the event types that this device supports as a mapping of
        supported event types to lists of handled event codes.

        Example
        --------
        >>> device.capabilities()
        { 1: [272, 273, 274],
          2: [0, 1, 6, 8] }

        If ``verbose`` is ``True``, event codes and types will be resolved
        to their names.

        ::

          { ('EV_KEY', 1): [('BTN_MOUSE', 272),
                            ('BTN_RIGHT', 273),
                            ('BTN_MIDDLE', 273)],
            ('EV_REL', 2): [('REL_X', 0),
                            ('REL_Y', 1),
                            ('REL_HWHEEL', 6),
                            ('REL_WHEEL', 8)] }

        Unknown codes or types will be resolved to ``'?'``.

        If ``absinfo`` is ``True``, the list of capabilities will also
        include absolute axis information in the form of
        :class:`AbsInfo` instances::

          { 3: [ (0, AbsInfo(min=0, max=255, fuzz=0, flat=0)),
                 (1, AbsInfo(min=0, max=255, fuzz=0, flat=0)) ]}

        Combined with ``verbose`` the above becomes::

          { ('EV_ABS', 3): [ (('ABS_X', 0), AbsInfo(min=0, max=255, fuzz=0, flat=0)),
                             (('ABS_Y', 1), AbsInfo(min=0, max=255, fuzz=0, flat=0)) ]}

        """

        if verbose:
            return dict(util.resolve_ecodes_dict(self._capabilities(absinfo)))
        else:
            return self._capabilities(absinfo)

    def input_props(self, verbose: bool = False):
        """
        Get device properties and quirks.

        Example
        -------
        >>> device.input_props()
        [0, 5]

        If ``verbose`` is ``True``, input properties are resolved to their
        names. Unknown codes are resolved to ``'?'``::

        [('INPUT_PROP_POINTER', 0), ('INPUT_PROP_POINTING_STICK', 5)]

        """
        props = _input.ioctl_EVIOCGPROP(self.fd)
        if verbose:
            return util.resolve_ecodes(ecodes.INPUT_PROP, props)

        return props

    def leds(self, verbose: bool = False):
        """
        Return currently set LED keys.

        Example
        -------
        >>> device.leds()
        [0, 1, 8, 9]

        If ``verbose`` is ``True``, event codes are resolved to their
        names. Unknown codes are resolved to ``'?'``::

        [('LED_NUML', 0), ('LED_CAPSL', 1), ('LED_MISC', 8), ('LED_MAIL', 9)]

        """
        leds = _input.ioctl_EVIOCG_bits(self.fd, ecodes.EV_LED)
        if verbose:
            return util.resolve_ecodes(ecodes.LED, leds)

        return leds

    def set_led(self, led_num: int, value: int) -> None:
        """
        Set the state of the selected LED.

        Example
        -------
        >>> device.set_led(ecodes.LED_NUML, 1)
        """
        self.write(ecodes.EV_LED, led_num, value)

    def __eq__(self, other):
        """
        Two devices are equal if their :data:`info` attributes are equal.
        """
        return isinstance(other, self.__class__) and self.info == other.info and self.path == other.path

    def __str__(self) -> str:
        msg = 'device {}, name "{}", phys "{}", uniq "{}"'
        return msg.format(self.path, self.name, self.phys, self.uniq or "")

    def __repr__(self) -> str:
        msg = (self.__class__.__name__, self.path)
        return "{}({!r})".format(*msg)

    def __fspath__(self):
        return self.path

    def close(self) -> None:
        if self.fd > -1:
            try:
                super().close()
                os.close(self.fd)
            finally:
                self.fd = -1

    def grab(self) -> None:
        """
        Grab input device using ``EVIOCGRAB`` - other applications will
        be unable to receive events until the device is released. Only
        one process can hold a ``EVIOCGRAB`` on a device.

        Warning
        -------
        Grabbing an already grabbed device will raise an ``OSError``.
        """

        _input.ioctl_EVIOCGRAB(self.fd, 1)

    def ungrab(self) -> None:
        """
        Release device if it has been already grabbed (uses `EVIOCGRAB`).

        Warning
        -------
        Releasing an already released device will raise an
        ``OSError('Invalid argument')``.
        """

        _input.ioctl_EVIOCGRAB(self.fd, 0)

    @contextlib.contextmanager
    def grab_context(self) -> Iterator[None]:
        """
        A context manager for the duration of which only the current
        process will be able to receive events from the device.
        """
        self.grab()
        yield
        self.ungrab()

    def upload_effect(self, effect: "ff.Effect"):
        """
        Upload a force feedback effect to a force feedback device.
        """

        data = memoryview(effect).tobytes()
        ff_id = _input.upload_effect(self.fd, data)
        return ff_id

    def erase_effect(self, ff_id) -> None:
        """
        Erase a force effect from a force feedback device. This also
        stops the effect.
        """

        _input.erase_effect(self.fd, ff_id)

    @property
    def repeat(self):
        """
        Get or set the keyboard repeat rate (in characters per
        minute) and delay (in milliseconds).
        """

        return KbdInfo(*_input.ioctl_EVIOCGREP(self.fd))

    @repeat.setter
    def repeat(self, value: Tuple[int, int]):
        return _input.ioctl_EVIOCSREP(self.fd, *value)

    def active_keys(self, verbose: bool = False):
        """
        Return currently active keys.

        Example
        -------

        >>> device.active_keys()
        [1, 42]

        If ``verbose`` is ``True``, key codes are resolved to their
        verbose names. Unknown codes are resolved to ``'?'``. For
        example::

          [('KEY_ESC', 1), ('KEY_LEFTSHIFT', 42)]

        """
        active_keys = _input.ioctl_EVIOCG_bits(self.fd, ecodes.EV_KEY)
        if verbose:
            return util.resolve_ecodes(ecodes.KEY, active_keys)

        return active_keys

    def absinfo(self, axis_num: int):
        """
        Return current :class:`AbsInfo` for input device axis

        Arguments
        ---------
        axis_num : int
          EV_ABS keycode (example :attr:`ecodes.ABS_X`)

        Example
        -------
        >>> device.absinfo(ecodes.ABS_X)
        AbsInfo(value=1501, min=-32768, max=32767, fuzz=0, flat=128, resolution=0)
        """
        return AbsInfo(*_input.ioctl_EVIOCGABS(self.fd, axis_num))

    def set_absinfo(self, axis_num: int, value=None, min=None, max=None, fuzz=None, flat=None, resolution=None) -> None:
        """
        Update :class:`AbsInfo` values. Only specified values will be overwritten.

        Arguments
        ---------
        axis_num : int
          EV_ABS keycode (example :attr:`ecodes.ABS_X`)

        Example
        -------
        >>> device.set_absinfo(ecodes.ABS_X, min=-2000, max=2000)

        You can also unpack AbsInfo tuple that will overwrite all values

        >>> device.set_absinfo(ecodes.ABS_Y, *AbsInfo(0, -2000, 2000, 0, 15, 0))
        """

        cur_absinfo = self.absinfo(axis_num)
        new_absinfo = AbsInfo(
            value if value is not None else cur_absinfo.value,
            min if min is not None else cur_absinfo.min,
            max if max is not None else cur_absinfo.max,
            fuzz if fuzz is not None else cur_absinfo.fuzz,
            flat if flat is not None else cur_absinfo.flat,
            resolution if resolution is not None else cur_absinfo.resolution,
        )
        _input.ioctl_EVIOCSABS(self.fd, axis_num, new_absinfo)
